//---------------------------------------------------------------
//---------------------------------------------------------------
//------------  Parameter-less SGA		             ------------
//------------  - function optimization				 ------------
//------------  - binary representation              ------------  
//------------  - generational replacement strategy  ------------  
//---------------------------------------------------------------
//---------------------------------------------------------------

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <conio.h>
#include <time.h>
#include <string.h>
#include <float.h>
//----------------
#include "general.h"
#include "individual.h"
#include "generation_sga.h"

//------- Parameter-less GA control parameters
int maxGenes;
int *popSize;				// population sizes of all running SGAs
int smallestPopSize;		// smallest population in the system
tind **populations;			// populations of all SGAs;
tind bestSolution;			// best solution of the run
int maxEvaluations;			// stopping criterion - overall maximal number of fitness evaluations cummulated over all SGAs
int currEvaluations;		// current number of calculated fitness function evaluations
int nSGA;					// number of currently running SGAs
int maxSGA;					// maximal number of running SGAs
int counter;				// global generation counter
int counterBase;			// the base of the global generation counter (referred to as the parameter 'm' in the articles)
int maxDigits;				// maximum number of digits of the counter number represented in number system of base counterBase
int *counterDigits;			// actual digits of the current counter number
tstat *statdata;			// array of population statistics for all SGAs
toverallstat overallStatData;

//------- Others
int rand_seed;			// seeds for pseudo-random generator

//---------------------------
char *name_output = "sga_func_opt.vys\0";  // output file
FILE *report_file;


//------------------------------------------------------------------------------------
//------- Increments the counter in the number system with base counterBase.
//------- Returns the highest digit that changed while incrementing the counter.
//------- Returns -1, if the counter exceeds the maxDigits.
int incrementCounter(){
	int i = 0;
	int carryOver = 1, pom;

	while((carryOver == 1) && (i < maxDigits)){
		pom = (counterDigits[i] + carryOver) / counterBase;
		counterDigits[i] = (counterDigits[i] + carryOver) % counterBase;
		carryOver = pom;
		i++;
	}
	counter++;

	if(i == maxDigits)
		return -1;
	else
		return i-1;
}

//----------------------------------------------------------------------
//------- Resets the generation counter
void resetCounter(){
	int i;

	counter = 0;				// global generation counter
	for(i=0; i<maxDigits; i++)
		counterDigits[i] = 0;	// reset the counter
	counterDigits[0] = -1;
}

//----------------------------------------------------------------------
//------- overall initialisation
//------- allocates memory for individuals
void initAll(){
	int i;

	//------- Random number generator seed
	srand(rand_seed = 2);

	//------- SGA parameters
	nTournament = 4;		// parameter of tournament selection
	pCross = 50;			// probability of crossover [%]

	//------- Initializes control parameters
	maxGenes = 80;					// fixed length of chromosomes
	maxEvaluations = 400000;		// stopping criterion - overall maximal number of fitness evaluations cummulated over all SGAs
	currEvaluations = 0;			// current number of calculated fitness function evaluations
	nSGA = 0;						// number of currently running SGAs
	maxSGA = 4;						// maximal number of running SGAs
	smallestPopSize = 64;
	counterBase = 6;				// the base of the global generation counter (referred to as the parameter 'm' in the articles)
	maxDigits = 10;					// maximum number of digits of the counter number represented in number system of base counterBase
	overallStatData.favg = -1.0e100;
	overallStatData.fbest = -1.0e100;
	overallStatData.popSize = 0;
	overallStatData.when = 0;

	//------- initializes counter digits
	counterDigits = (int*)malloc(maxDigits * sizeof(int));			// actual digits of the current counter number
	resetCounter();			// reset the counter

	//------- allocates memory for pointers to maxSGA populations
	populations = (tind **)malloc(maxSGA * sizeof(tind*));
	for(i=0; i<maxSGA; i++)
		populations[i] = NULL;

	bestSolution.genes = (int*)malloc(maxGenes * sizeof(int));

	//------- allocates an array of population sizes
	popSize = (int*)malloc(maxSGA * sizeof(int));	// population sizes of all running SGAs

	//------- allocates memory for an array of statdata records
	statdata = (tstat*)malloc(maxSGA * sizeof(tstat));			// array of population statistics for all SGAs
}

//----------------------------------------------------------------------
//------- Releases all of the allocated memory
void freeAll(){
	int i;

	//------- releases memory allocated for pointers to maxSGA populations
	for(i=0; i<nSGA; i++)
		destroyPopulation(i);
	free(populations);

	//------- releases memory allocated for an array of population sizes
	free(popSize);
	free(bestSolution.genes);

	//------- releases memory allocated for an array of statdata records
	free(statdata);

	//------- releases the counter
	free(counterDigits);			// actual digits of the current counter number
}

//----------------------------------------------------------------------
//------- Destroys the population 'popId' and all populations that are smaller than 'popId'
void destroySmallerPopulationsThan(int popId){
	int i, j;

	for(i=0; i<=popId; i++){
		printf("\n    !!! Population %d destroyed:\t fitness=%.3f (%.3f)", popSize[0], statdata[0].favg, statdata[0].fbest);
		destroyPopulation(0);			// destroy population popId
		for(j=1; j<maxSGA; j++)
			shiftPopulationToLeft(j);	// shifts larger populations to the left
		nSGA--;							// decrement nSGA
	}
	resetCounter();						// reset counter
}

//----------------------------------------------------------------------
//------- Destroys the population 'popId'
void destroyConvergedPopulation(int popId){
	int i;

	printf("\n    !!! Population %d destroyed:\t fitness=%.3f (%.3f)", popSize[popId], statdata[popId].favg, statdata[popId].fbest);
	destroyPopulation(popId);		// destroy population popId
	for(i=popId+1; i<maxSGA; i++)
		shiftPopulationToLeft(i);	// shifts larger populations to the left
	nSGA--;							// decrement nSGA
	resetCounter();					// reset counter
}

//----------------------------------------------------------------------
//------- Final report
void finalReport(){
	int i;

	printf("\n\n======= FINAL REPORT =======");
	printf("\nTotal number of fitness evaluations: %d", currEvaluations);
	printf("\nBest-of-the-run solution: %.3f", overallStatData.fbest);
	printf("\nFound in population of size: %d", overallStatData.popSize);
	printf("\nNumber of FE needed to find it: %d", overallStatData.when);
	printf("\nThe best solution found: ");
	for(i=0; i<maxGenes; i++)
		printf("%d", bestSolution.genes[i]);

	printf("\n\nActive populations:");
	for(i=0; i<nSGA; i++)
		printf("\n%d:\tpopSize=%d, avgFitness=%.3f, bestFitness=%.3f", i, popSize[i], statdata[i].favg, statdata[i].fbest);
}

//--------------------------------------------------------
//-------------------- MAIN EVOLUTION --------------------
//--------------------------------------------------------
int main(int argc,char *argv[]){
	int i, j;
	int par1, par2;				// indexes of parents
	int individualsGenerated;   // a number of newly generated individuals

	//-------
	report_file = fopen(name_output,"w"); 

	//------- Initialization
	initAll();

	//-----------------------------------------------
	//------- Main loop of the parameter-less SGA
	//-----------------------------------------------
	do{//------- generate new population
		//------ 1. Increment the iteration counter and choose the SGA which will generate a new population in this iteration.
		if((i = incrementCounter()) == -1){
			printf("\nError: Counter overflow");
			return 1;
		}
		//------ Perform one generation with SGA with index 'i'
		else{
			if(populations[i] == NULL){
				//------- initialization of the first population
				if(initPopulation(i, &currEvaluations) == 0)
					return 1;
				if(overallStatData.fbest < statdata[i].fbest){
					overallStatData.fbest = statdata[i].fbest;
					overallStatData.favg = statdata[i].favg;
					overallStatData.popSize = popSize[i];
					overallStatData.when = currEvaluations;
					for(j=0; j<maxGenes; j++)
						bestSolution.genes[j] = populations[i][statdata[i].indbest].genes[j];
				}
				nSGA++;	// increment the number of currently running SGAs
			}
			else{
				//------- Perform one generation with the population popId
				doGeneration(i, &currEvaluations);
				if(overallStatData.fbest < statdata[i].fbest){
					overallStatData.fbest = statdata[i].fbest;
					overallStatData.favg = statdata[i].favg;
					overallStatData.popSize = popSize[i];
					overallStatData.when = currEvaluations;
					for(j=0; j<maxGenes; j++)
						bestSolution.genes[j] = populations[i][statdata[i].indbest].genes[j];
				}
				//------ 2. Check the convergence of the current SGA 
				//-------   and destroy its population if needed.
				if(checkConvergence(i) == 1){
					destroyConvergedPopulation(i);
				}
			}
		}

		//------ 3. Check the relation regarding the average fitness values of subsequent populations 
		//-------   and destroy some smaller populations if needed.
		if((i = checkAverageFitness(statdata)) > -1){
			destroyConvergedPopulation(i);
//			destroySmallerPopulationsThan(i);
		}

	}while(currEvaluations < maxEvaluations);

	//------- Final report
	finalReport();

	//------- Cleaning
	freeAll();

	//-------
	printf("\n\nFinished ...");
	getch();
	return 0;
}
